@@ -40,7 +40,8 @@ Quadtree_init(Quadtree *self, PyObject *args, PyObject *kwds)
 static PyObject *
 Quadtree_add(Quadtree *self, PyObject *args)
 {
-    int n;
+    int n, size;
+    PyObject *bounds=NULL;
     double min[2], max[2];
     SHPObject *s;
 
@@ -48,10 +49,37 @@ Quadtree_add(Quadtree *self, PyObject *args)
     int part_start[1] = {0};
     double x[5], y[5];
     
-    if (!PyArg_ParseTuple(args, "i(dddd)", &n,
-                          &min[0], &min[1], &max[0], &max[1]))
+    if (!PyArg_ParseTuple(args, "iO", &n, &bounds))
         return NULL;
 
+    /* Check length of the bounds argument */
+    if (!PySequence_Check(bounds))
+    {
+        PyErr_SetString(PyExc_TypeError, "Bounds must be a sequence");
+        return NULL;
+    }
+
+    size = (int) PySequence_Size(bounds);
+    if (size == 2)
+    {
+        min[0] = max[0] = PyFloat_AsDouble(PySequence_ITEM(bounds, 0));
+        min[1] = max[1] = PyFloat_AsDouble(PySequence_ITEM(bounds, 1));
+    }
+    else if (size == 4)
+    {
+        min[0] = PyFloat_AsDouble(PySequence_ITEM(bounds, 0));
+        min[1] = PyFloat_AsDouble(PySequence_ITEM(bounds, 1));
+        max[0] = PyFloat_AsDouble(PySequence_ITEM(bounds, 2));
+        max[1] = PyFloat_AsDouble(PySequence_ITEM(bounds, 3));
+    }
+    else
+    {
+        PyErr_Format(PyExc_TypeError,
+            "Bounds argument must be sequence of length 2 or 4, not %d",
+            size);
+        return NULL;
+    }
+    
     /* make shape vertices */
     x[0] = min[0];
     x[1] = min[0];
@@ -78,6 +106,59 @@ Quadtree_add(Quadtree *self, PyObject *args)
     return Py_None;
 }
 
+static PyObject *
+Quadtree_prune(Quadtree *self)
+{
+    SHPTreeTrimExtraNodes(self->tree);
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+/* Tree structure */
+
+static void 
+QuadtreeNodeDump(SHPTreeNode *node, PyObject *nodelist)
+{
+    int i;
+    PyObject *g=NULL, *ids=NULL, *bounds=NULL, *children=NULL;
+
+    g = PyDict_New();
+    
+    /* Save bounds */
+    bounds = Py_BuildValue("(dddd)", 
+                node->adfBoundsMin[0], node->adfBoundsMin[1], 
+                node->adfBoundsMax[0], node->adfBoundsMax[1]); 
+    PyDict_SetItemString(g, "bounds", bounds);
+ 
+    /* Save ids */
+    ids = PyList_New(0);
+    for (i=0; i<node->nShapeCount; i++)
+    {
+        PyList_Append(ids, Py_BuildValue("i", node->panShapeIds[i]));
+    }
+    PyDict_SetItemString(g, "ids", ids);
+
+    /* Child nodes */
+    children = PyList_New(0);
+    for (i=0; i<node->nSubNodes; i++)
+    {
+        if (node->apsSubNode[i] != NULL)
+            QuadtreeNodeDump(node->apsSubNode[i], children);
+    }
+    PyDict_SetItemString(g, "nodes", children);
+    
+    /* Add g to the node list */
+    PyList_Append(nodelist, g);
+}
+
+static PyObject *
+Quadtree_struct(Quadtree *self)
+{
+    PyObject *nodes=NULL;
+    nodes = PyList_New(0);
+    QuadtreeNodeDump(self->tree->psRoot, nodes);
+    return PyList_GetItem(nodes, 0);
+}
 
 static PyObject *
 Quadtree_intersection(Quadtree *self, PyObject *args)
@@ -103,6 +184,8 @@ Quadtree_intersection(Quadtree *self, PyObject *args)
 
 static PyMethodDef module_methods[] = {
     {"add", (PyCFunction)Quadtree_add, METH_VARARGS, "Add an item to an index, specifying an integer id and a bounding box"},
+    {"prune", (PyCFunction)Quadtree_prune, METH_NOARGS, "Remove unused nodes from the tree"},
+    {"struct", (PyCFunction)Quadtree_struct, METH_NOARGS, "Return simple graph structure of the tree"},
     {"likely_intersection", (PyCFunction)Quadtree_intersection, METH_VARARGS, "Return an iterator over integer ids of items that are likely to intersect with the specified bounding box."},
     {NULL}
 };
